# generated by datamodel-codegen:
#   filename:  SourceFileInfo.yaml

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel


class SourceFileInfo(BaseModel):
    metadata_etag: Optional[str] = None
    metadata_file_path: Optional[str] = None
    metadata_bucket_name: Optional[str] = None
    metadata_last_modified: Optional[str] = None
    registry_entry_generated_at: Optional[str] = None
