# generated by datamodel-codegen:
#   filename:  ConnectorIPCOptions.yaml

from __future__ import annotations

from typing import List

from pydantic import BaseModel, Extra
from typing_extensions import Literal


class DataChannel(BaseModel):
    class Config:
        extra = Extra.forbid

    version: str
    supportedSerialization: List[Literal["JSONL", "PROTOBUF", "FLATBUFFERS"]]
    supportedTransport: List[Literal["STDIO", "SOCKET"]]


class ConnectorIPCOptions(BaseModel):
    class Config:
        extra = Extra.forbid

    dataChannel: DataChannel
