# generated by datamodel-codegen:
#   filename:  ResourceRequirements.yaml

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Extra


class ResourceRequirements(BaseModel):
    class Config:
        extra = Extra.forbid

    cpu_request: Optional[str] = None
    cpu_limit: Optional[str] = None
    memory_request: Optional[str] = None
    memory_limit: Optional[str] = None
