# generated by datamodel-codegen:
#   filename:  ConnectorMetrics.yaml

from __future__ import annotations

from typing import Any, Optional, Union

from pydantic import BaseModel, Extra
from typing_extensions import Literal


class ConnectorMetrics(BaseModel):
    all: Optional[Any] = None
    cloud: Optional[Any] = None
    oss: Optional[Any] = None


class ConnectorMetric(BaseModel):
    class Config:
        extra = Extra.allow

    usage: Optional[Union[str, Literal["low", "medium", "high"]]] = None
    sync_success_rate: Optional[Union[str, Literal["low", "medium", "high"]]] = None
    connector_version: Optional[str] = None
