# generated by datamodel-codegen:
#   filename:  SecretStore.yaml

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Extra, Field
from typing_extensions import Literal


class SecretStore(BaseModel):
    class Config:
        extra = Extra.forbid

    alias: Optional[str] = Field(
        None,
        description="The alias of the secret store which can map to its actual secret address",
    )
    type: Optional[Literal["GSM"]] = Field(
        None, description="The type of the secret store"
    )
