# generated by datamodel-codegen:
#   filename:  ConnectorPackageInfo.yaml

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel


class ConnectorPackageInfo(BaseModel):
    cdk_version: Optional[str] = None
